/*
 Yliluoma's ordered dithering algorithm 1 a
 Usage: ./yliluoma_ordered_dither_1a.exe scene.png scenedither1.png

 Arbitrary-palette positional dithering algorithm
 https://bisqwit.iki.fi/story/howto/dither/jy/
 */

#include <gd.h>
#include <stdio.h>

/* 8x8 threshold map */
#define d(x) x / 64.0
static const double map[8 * 8] = {
    d(0), d(48), d(12), d(60), d(3), d(51), d(15), d(63),
    d(32), d(16), d(44), d(28), d(35), d(19), d(47), d(31),
    d(8), d(56), d(4), d(52), d(11), d(59), d(7), d(55),
    d(40), d(24), d(36), d(20), d(43), d(27), d(39), d(23),
    d(2), d(50), d(14), d(62), d(1), d(49), d(13), d(61),
    d(34), d(18), d(46), d(30), d(33), d(17), d(45), d(29),
    d(10), d(58), d(6), d(54), d(9), d(57), d(5), d(53),
    d(42), d(26), d(38), d(22), d(41), d(25), d(37), d(21)};
#undef d

/* Palette */
static const unsigned pal[16] =
    {0x080000, 0x201A0B, 0x432817, 0x492910,
     0x234309, 0x5D4F1E, 0x9C6B20, 0xA9220F,
     0x2B347C, 0x2B7409, 0xD0CA40, 0xE8A077,
     0x6A94AB, 0xD5C4B3, 0xFCE76E, 0xFCFAE2};

// Compare the difference of two RGB values
double ColorCompare(int r1, int g1, int b1, int r2, int g2, int b2)
{
    double diffR = (r1 - r2) / 255.0, diffG = (g1 - g2) / 255.0, diffB = (b1 - b2) / 255.0;
    return diffR * diffR + diffG * diffG + diffB * diffB;
}

double EvaluateMixingError(int r, int g, int b,    // Desired color
                           int r0, int g0, int b0, // Mathematical mix product
                           int r1, int g1, int b1, // Mix component 1
                           int r2, int g2, int b2, // Mix component 2
                           double ratio)           // Mixing ratio
{
    return ColorCompare(r, g, b, r0, g0, b0);
}

struct MixingPlan
{
    unsigned colors[2];
    double ratio; /* 0 = always index1, 1 = always index2, 0.5 = 50% of both */
};

MixingPlan DeviseBestMixingPlan(unsigned color)
{
    const unsigned r = color >> 16, g = (color >> 8) & 0xFF, b = color & 0xFF;
    MixingPlan result = {{0, 0}, 0.5};
    double least_penalty = 1e99;
    // Loop through every unique combination of two colors from the palette,
    // and through each possible way to mix those two colors. They can be
    // mixed in exactly 64 ways, when the threshold matrix is 8x8.
    for (unsigned index1 = 0; index1 < 16; ++index1)
        for (unsigned index2 = index1; index2 < 16; ++index2)
            for (unsigned ratio = 0; ratio < 64; ++ratio)
            {
                if (index1 == index2 && ratio != 0)
                    break;
                // Determine the two component colors
                unsigned color1 = pal[index1], color2 = pal[index2];
                unsigned r1 = color1 >> 16, g1 = (color1 >> 8) & 0xFF, b1 = color1 & 0xFF;
                unsigned r2 = color2 >> 16, g2 = (color2 >> 8) & 0xFF, b2 = color2 & 0xFF;

                // Determine what mixing them in this proportion will produce
                unsigned r0 = r1 + ratio * int(r2 - r1) / 64;
                unsigned g0 = g1 + ratio * int(g2 - g1) / 64;
                unsigned b0 = b1 + ratio * int(b2 - b1) / 64;

                // Determine how well that matches what we want to accomplish
                double penalty = EvaluateMixingError(
                    r, g, b, r0, g0, b0, r1, g1, b1, r2, g2, b2,
                    ratio / 64.0);

                if (penalty < least_penalty)
                {
                    // Keep the result that has the smallest error
                    least_penalty = penalty;
                    result.colors[0] = index1;
                    result.colors[1] = index2;
                    result.ratio = ratio / 64.0;
                }
            }
    return result;
}

int main(int argc, char **argv)
{
    if (argc != 3)
    {
        printf("Usage: %s INPUT.png OUTPUT.png\n", argv[0]);
        return 0;
    }

    FILE *fp = fopen(argv[1], "rb");
    gdImagePtr srcim = gdImageCreateFromPng(fp);
    fclose(fp);

    unsigned w = gdImageSX(srcim), h = gdImageSY(srcim);
    gdImagePtr im = gdImageCreate(w, h);
    for (unsigned c = 0; c < 16; ++c)
        gdImageColorAllocate(im, pal[c] >> 16, (pal[c] >> 8) & 0xFF, pal[c] & 0xFF);
#pragma omp parallel for
    for (unsigned y = 0; y < h; ++y)
        for (unsigned x = 0; x < w; ++x)
        {
            double map_value = map[(x & 7) + ((y & 7) << 3)];
            unsigned color = gdImageGetTrueColorPixel(srcim, x, y);
            MixingPlan plan = DeviseBestMixingPlan(color);
            gdImageSetPixel(im, x, y, plan.colors[map_value < plan.ratio ? 1 : 0]);
        }
    fp = fopen(argv[2], "wb");
    gdImagePng(im, fp);
    fclose(fp);
    gdImageDestroy(im);
    gdImageDestroy(srcim);
}
